//
//  Gradient.m
//  GradientDemo
//
//  Created by Guanshan Liu on 31/07/2012.
//  Copyright (c) 2012 Guanshan Liu. All rights reserved.
//

#import "Gradient.h"

#define kGradientColorsKey      @"colors"
#define kGradientLocationsKey   @"locations"

#define LOCATION_IS_VALID(x)    ((x) >= 0.0 && (x) <= 1.0)
#define DEGREES_TO_RADIANS(i)   (M_PI * (i) / 180.0)

@implementation Gradient

#pragma mark - NSCoding

- (id)initWithCoder:(NSCoder *)aDecoder {
    self = [super init];
    if (self) {
        _colors = (__bridge CFMutableArrayRef)([aDecoder decodeObjectForKey:kGradientColorsKey]);
        NSArray *locationArray = [aDecoder decodeObjectForKey:kGradientLocationsKey];
        _locations = malloc(locationArray.count * sizeof(CGFloat));
        [locationArray enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop){
            _locations[idx] = [obj floatValue];
        }];
        _colorSpace = CGColorSpaceCreateDeviceRGB();
    }
    return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder {
    [aCoder encodeObject:(__bridge id)(_colors) forKey:kGradientColorsKey];
    int count = CFArrayGetCount(_colors);
    NSMutableArray *locationArray = [NSMutableArray arrayWithCapacity:count];
    for (int i = 0; i < count; ++i) {
        [locationArray addObject:@( _locations[i] )];
    }
    [aCoder encodeObject:locationArray forKey:kGradientLocationsKey];
}

#pragma mark - NSCopying

- (id)copyWithZone:(NSZone *)zone {
    Gradient *copy = [[Gradient allocWithZone:zone] init];
    if (copy) {
        copy->_colors = self->_colors;
        copy->_locations = self->_locations;
        copy->_colorSpace = self->_colorSpace;
    }
    return copy;
}

#pragma mark - Auxiliary Functions

- (void)setGradientLocations {
    int count = CFArrayGetCount(_colors);
    _locations = malloc(sizeof(CGFloat) * count);
    for (int i = 0; i < count; ++i) {
        _locations[i] = (i) ? ((float)i / (float)(count - 1)) : 0.0f;
    }
}

#pragma mark - Initialization

- (id)initWithStartingColor:(UIColor *)startingColor endingColor:(UIColor *)endingColor {
    self = [super init];
    if (self) {
        _colors = CFArrayCreateMutable(kCFAllocatorDefault, 2, &kCFTypeArrayCallBacks);
        CFArrayAppendValue(_colors, startingColor.CGColor);
        CFArrayAppendValue(_colors, endingColor.CGColor);
        _locations = malloc(sizeof(CGFloat) * 2);
        _locations[0] = 0.0f;
        _locations[1] = 1.0f;
        _colorSpace = CGColorSpaceCreateDeviceRGB();
    }
    return self;
}

- (id)initWithColors:(NSArray *)colorArray {
    self = [super init];
    if (self) {
        _colors = CFArrayCreateMutable(kCFAllocatorDefault, colorArray.count, &kCFTypeArrayCallBacks);
        for (UIColor* color in colorArray) {
            CFArrayAppendValue(_colors, color.CGColor);
        };
        [self setGradientLocations];
        _colorSpace = CGColorSpaceCreateDeviceRGB();
    }
    return self;
}

- (id)initWithColorsAndLocations:(UIColor *)firstColor, ... {
    self = [super init];
    if (self) {
        _colors = CFArrayCreateMutable(kCFAllocatorDefault, 0, &kCFTypeArrayCallBacks);
        NSMutableArray *locationArray = [NSMutableArray array];
        
        va_list arguments;
        va_start(arguments, firstColor);
        UIColor *color = firstColor;
        while (color) {
            CGFloat loc = va_arg(arguments, double);
            if (!LOCATION_IS_VALID(loc)) {
                break;
            }
            CFArrayAppendValue(_colors, color.CGColor);
            [locationArray addObject:@( loc )];
            color = va_arg(arguments, UIColor *);
        }
        va_end(arguments);
        
        _locations = malloc(sizeof(CGFloat) * locationArray.count);
        [locationArray enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop){
            _locations[idx] = [obj floatValue];
        }];
        _colorSpace = CGColorSpaceCreateDeviceRGB();
    }
    return self;
}

- (id)initWithColors:(NSArray *)colorArray atLocations:(const CGFloat *)locations colorSpace:(CGColorSpaceRef)colorSpace {
    self = [super init];
    if (self) {
        int count = colorArray.count;
        _colors = CFArrayCreateMutable(kCFAllocatorDefault, count, &kCFTypeArrayCallBacks);
        for (UIColor* color in colorArray) {
            CFArrayAppendValue(_colors, color.CGColor);
        };
        
        _locations = malloc(sizeof(CGFloat) * count);
        for (int i = 0; i < count; ++i) {
            _locations[i] = locations[i];
        }
        
        _colorSpace = colorSpace;
    }
    return self;
}

#pragma mark - Memory Management

- (void)dealloc {
    CFRelease(_colors);
    free(_locations);
    CGColorSpaceRelease(_colorSpace);
}

#pragma mark - Primitive Drawing Methods

- (void)drawFromPoint:(CGPoint)startingPoint toPoint:(CGPoint)endingPoint options:(GradientDrawingOptions)options {
    CGContextRef context = UIGraphicsGetCurrentContext();
    
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawLinearGradient(context, gradient, startingPoint, endingPoint, options);
    
    CGGradientRelease(gradient);
}

- (void)drawFromCenter:(CGPoint)startCenter radius:(CGFloat)startRadius toCenter:(CGPoint)endCenter radius:(CGFloat)endRadius options:(GradientDrawingOptions)options {
    CGContextRef context = UIGraphicsGetCurrentContext();
    
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawRadialGradient(context, gradient, startCenter, startRadius, endCenter, endRadius, options);
    
    CGGradientRelease(gradient);
}

#pragma mark - Drawing Linear Gradients

- (void)drawInRect:(CGRect)rect angle:(CGFloat)angle {
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSaveGState(context);
    CGContextClipToRect(context, rect);
    CGContextRotateCTM(context, -DEGREES_TO_RADIANS(angle));
    
    CGPoint startPoint = CGPointMake(CGRectGetMinX(rect), CGRectGetMidY(rect));
    CGPoint endPoint = CGPointMake(CGRectGetMaxX(rect), CGRectGetMidY(rect));
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawLinearGradient(context, gradient, startPoint, endPoint, kCGGradientDrawsAfterEndLocation|kCGGradientDrawsBeforeStartLocation);
    
    CGGradientRelease(gradient);
    CGContextRestoreGState(context);
}

- (void)drawInBezierPath:(UIBezierPath *)path angle:(CGFloat)angle; {
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSaveGState(context);
    [path addClip];
    CGContextClip(context);
    CGContextRotateCTM(context, -DEGREES_TO_RADIANS(angle));
    
    CGRect bounds = path.bounds;
    CGPoint startPoint = CGPointMake(CGRectGetMinX(bounds), CGRectGetMidY(bounds));
    CGPoint endPoint = CGPointMake(CGRectGetMaxX(bounds), CGRectGetMidY(bounds));
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawLinearGradient(context, gradient, startPoint, endPoint, kCGGradientDrawsAfterEndLocation|kCGGradientDrawsBeforeStartLocation);
    
    CGGradientRelease(gradient);
    CGContextRestoreGState(context);
}

#pragma mark - Drawing Radial Gradients

- (void)drawInRect:(CGRect)rect relativeCenterPosition:(CGPoint)relativeCenterPosition {
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSaveGState(context);
    CGContextClipToRect(context, rect);
    
    CGFloat width = CGRectGetWidth(rect);
    CGFloat height = CGRectGetHeight(rect);
    CGFloat radius = sqrtf(powf(width/2, 2) + powf(height/2, 2));
    CGPoint startCenter = CGPointMake(width/2 + (width * relativeCenterPosition.x)/2, height/2 + (height * relativeCenterPosition.y)/2);
    CGPoint endCenter = CGPointMake(width/2, height/2);
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawRadialGradient(context, gradient, startCenter, 0, endCenter, radius, 0);
    
    CGGradientRelease(gradient);
    CGContextRestoreGState(context);
}

- (void)drawInBezierPath:(UIBezierPath *)path relativeCenterPosition:(CGPoint)relativeCenterPosition {
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSaveGState(context);
    [path addClip];
    CGContextClip(context);
    
    CGRect bounds = path.bounds;
    CGFloat width = CGRectGetWidth(bounds);
    CGFloat height = CGRectGetHeight(bounds);
    CGFloat radius = sqrtf(powf(width/2, 2) + powf(height/2, 2));
    CGPoint startCenter = CGPointMake(width/2 + (width * relativeCenterPosition.x)/2, height/2 + (height * relativeCenterPosition.y)/2);
    CGPoint endCenter = CGPointMake(width/2, height/2);
    CGGradientRef gradient = CGGradientCreateWithColors(_colorSpace, _colors, _locations);
    CGContextDrawRadialGradient(context, gradient, startCenter, 0, endCenter, radius, 0);
    
    CGGradientRelease(gradient);
    CGContextRestoreGState(context);
}

#pragma mark - Getting Gradient Properties

- (CGColorSpaceRef)colorSpace {
    return _colorSpace;
}

- (NSInteger)numberOfColorStops {
    return CFArrayGetCount(_colors);
}

- (void)getColor:(UIColor **)color location:(CGFloat *)location atIndex:(NSInteger)index {
    if (index >= 0 && index < CFArrayGetCount(_colors)) {
        *color = [[UIColor alloc] initWithCGColor:(CGColorRef)CFArrayGetValueAtIndex(_colors, index)];
        *location = _locations[index];
    }
    else {
        *color = nil;
        *location = CGFLOAT_MAX;
    }
}

// TODO: cleanup
- (UIColor *)interpolatedColorAtLocation:(CGFloat)location {
    if (LOCATION_IS_VALID(location)) {
        if (location == 0.0f) {
            return [[UIColor alloc] initWithCGColor:(CGColorRef)CFArrayGetValueAtIndex(_colors, 0)];
        }
        
        int index;
        int count = CFArrayGetCount(_colors);
        for (index = 1; index < count; ++index) {
            if (location == _locations[index]) {
                return [[UIColor alloc] initWithCGColor:(CGColorRef)CFArrayGetValueAtIndex(_colors, index)];
            }
            else if (location < _locations[index]) {
                break;
            }
        }
        
        UIColor *startColor = [[UIColor alloc] initWithCGColor:(CGColorRef)CFArrayGetValueAtIndex(_colors, index - 1)];
        UIColor *endColor = [[UIColor alloc] initWithCGColor:(CGColorRef)CFArrayGetValueAtIndex(_colors, index)];
        CGFloat step = (location - _locations[index - 1]) / (_locations[index] - _locations[index - 1]);
        CGFloat r, g, b, a, r1, r2, g1, g2, b1, b2, a1, a2;
        [startColor getRed:&r1 green:&g1 blue:&b1 alpha:&a1];
        [endColor getRed:&r2 green:&g2 blue:&b2 alpha:&a2];
        r = r1 + (r2 - r1) * step;
        g = g1 + (g2 - g1) * step;
        b = b1 + (b2 - b1) * step;
        a = a1 + (a2 - a1) * step;
        return [UIColor colorWithRed:r green:g blue:b alpha:a];
    }
    
    return nil;
}

@end
